//  This file is part of par2cmdline (a PAR 2.0 compatible file verification and
//  repair tool). See http://parchive.sourceforge.net for details of PAR 2.0.
//
//  par2cmdline is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  par2cmdline is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "par2cmdline.h"

unsigned PrCap = 0;

// Sets PrCap to 1 for non-MMX processor, 2 for MMX processor.
// Only needs to be called once immediately after parsing command line.

void SetPrCap(const CommandLine &commandline) {
  if (PrCap == 0) {
    _asm {
	pushfd				;try to flip "ID" bit in eflags
	pop     eax
	btc     eax,21
	push    eax
	popfd
	pushfd
	pop	ecx
	xor     ecx,eax			;bit 21 == 0 if "ID" bit changed
	xor	eax,eax
	bt      ecx,21
	jc	fset			;jmp= no CPUID, no advanced opcodes
	cpuid				;eax= highest supported CPUID function
	test	eax,eax
	jz	fset			;jmp= can't get features
	mov	eax,1			;get features
	cpuid
	xor	eax,eax
	test	edx,1			;floating point?
	jz	fset			;jmp= no, no MMX
	bt	edx,23			;MMX support?
	jnc	fset			;jmp= nope
	inc	eax			;indicate MMX
fset:
	inc	eax
	mov	PrCap,eax
fdone:
    }
    if (PrCap < 2
	&& commandline.GetNoiseLevel() > CommandLine::nlQuiet
    ) {
	cout << "*** Warning: non-MMX processor ***" << endl;
    }
  }
}

/*----------------------------------------------------------------------------
;
; Paul Houle (paulhoule.com) March 7, 2008
; The below routine is the gist of the changes.  It is assembly code to
; handle the main CPU work of par2 creation.
;
;---------------------------------------------------------------------------*/

#pragma warning(disable:4731)		// inhibit VC's "ebp modified" warning

void AsmRSPrc(size_t bsize, u32 inputindex, const void *inputbuffer,
  u32 outputindex, void *outputbuffer, unsigned factor, void *table) {

  _asm {
	mov	eax,esp			;copy original esp
	and	esp,not 0fh		;para align the stack
	push	edx			;(these pushes are to maintain
	push	edx			;  a para boundary for the tables)
	push	eax			;save original esp
	push	ebp			;save original ebp
	sub	esp,800h		;[esp]= L, [esp+4]= H (interleaved)

	#define L [esp]
	#define H [esp+400h]
	#define SFRM [esp+800h]		;for access to ebp (stack frame ptr)

	mov	eax,factor
	movzx	ebx,ah
	movzx	eax,al
	shl	ebx,8
	shl	eax,8
	mov	edx,table
	lea	esi,[edx+eax*2]			;LL
	lea	ecx,[edx+ebx*2+(1*256*256)*2]	;HL
	lea	edi,[edx+eax*2+(1*256*256)*2]	;LH
	lea	edx,[edx+ebx*2+(2*256*256)*2]	;HH

	xor	ebx,ebx
	pxor	mm7,mm7
	align	16
LHmxlp:
	movq	mm0,[esi+ebx]
	pxor	mm0,[ecx+ebx]		;L3,L2,L1,L0
	movq	mm1,mm0
	punpcklwd mm0,mm7
	punpckhwd mm1,mm7
	movq	L[ebx*2],mm0
	movq	L[ebx*2+8],mm1

	movq	mm2,[edi+ebx]
	pxor	mm2,[edx+ebx]		;H3,H2,H1,H0
	movq	mm3,mm2
	punpcklwd mm2,mm7
	punpckhwd mm3,mm7
	movq	H[ebx*2],mm2
	movq	H[ebx*2+8],mm3

	add	ebx,8
	cmp	ebx,512
	jne	LHmxlp

	mov	esi,inputbuffer		;source
	mov	edi,outputbuffer	;destination
	mov	ebx,bsize		;src bytes to process (multiple of 4)

	or	ebx,ebx
	jz	eempt			;jmp= no data (shouldn't happen!)

	; At this point (chk8) ebx MUST be non-zero and a multiple of four
chk8:
	; Because of source read-ahead and alignment issues, don't bother
	; trying to use the 8-byte-per MMX loop unless the block is sizeable

	cmp	ebx,120			;worth bothering with the MMX code?
	jb	lp4			;jmp= no, handle in dwords

	; esi and edi are always relatively 8-byte aligned, but the calling
	; code operates in dword blocks so they ARE occasionally not
	; physically 8-byte aligned.  Misalignment slows things down quite a
	; bit; so, if the esi/edi addresses aren't 8-byte aligned, process a
	; single dword beforehand to make them so.

	mov	eax,ebx			;save true length
	mov	ebx,4			;assume we'll do a dword to align
	test	edi,ebx			;test for qword misalignment
	jnz	lp4			;jmp= misaligned, do a dword to align

	; esi & edi are qword aligned and we have a decent chunk of data, so
	; it's worthwhile to use the MMX loop.  First round down the size to
	; a multiple of 8 so we can sense when we're done.  Also we reduce
	; the processing amount by 8 since we read-ahead -- if we didn't,
	; we'd read beyond the end of the input buffer.

	mov	ebx,eax			;restore true length
	and	ebx,not 4		;only do a multiple of qwords below
	sub	ebx,8			;limit further due to read-ahead

	mov	eax,[esi]		;load 1st 8 source bytes
	movd	mm4,[esi+4]

	add	esi,ebx			;point to end of input/output
	add	edi,ebx
	neg	ebx			;convert byte size to count-up

	; The below loop processes 8 bytes at a time.  It runs at approx.
	; 10.5 clocks per iteration on a Core 2 Duo.
	; That's close to optimal -- which is 10 -- since there are 10 mem
	; reads in the loop (there's only one mem read port in the
	; architecture).  The extra .5 clock appears to be a
	; cache-loading delay, since if the read addresses below are made
	; invariant it goes away.
	;
	; The 8 byte MMX loop (lp8) runs 20-30% faster than the simple four
	; byte loop (lp4) on a Core 2 Duo.  The speedup is really only due to
	; lp8 doing less memory fetches (8 bytes per srd/dst instead of 4).
	;
	; Some work was done to achieve U/V pipe pairing for early Pentiums,
	; but not at the expense of the Core 2 Duo performance.

	align	16
lp8:
	movzx	edx,al
	movzx	ecx,ah
	shr	eax,16
	movd	mm0,L[edx*4]
	movd	mm1,H[ecx*4]
	movzx	edx,al
	movzx	ecx,ah
	movd	eax,mm4
	movq	mm4,[esi+ebx+8]		;read-ahead next 8 source bytes
	movd	mm2,L[edx*4]
	movzx	edx,al
	movq	mm5,[edi+ebx]
	movd	mm3,H[ecx*4]
	movzx	ecx,ah
	shr	eax,16
	punpckldq mm0,L[edx*4]
	movzx	edx,al
	punpckldq mm1,H[ecx*4]
	movzx	ecx,ah
	punpckldq mm2,L[edx*4]
	pxor	mm1,mm0
	punpckldq mm3,H[ecx*4]
	pxor	mm3,mm2
	pslld	mm3,16
	movd	eax,mm4			;prepare src bytes 3-0 for next loop
	pxor	mm1,mm5
	psrlq	mm4,32			;align src bytes 7-4 for next loop
	pxor	mm1,mm3
	movq	[edi+ebx],mm1
	add	ebx,8
	jnz	lp8
	jmp short ewrap

	; The below loop handles single dwords to align the src/dst prior
	; to processing 8-byte (mmx loop) blocks, or it handles dwords that
	; can't be processed by the mmx loop.
lp4:
	mov	eax,[esi]
	add	esi,4
	mov	ecx,eax
	shr	eax,16
	movzx	edx,al
	shr	eax,8
	mov	edx,L[edx*4]		;look up 2nd solution
	xor	edx,H[eax*4]
	movzx	eax,ch
	movzx	ecx,cl
	shl	edx,16
	mov	eax,H[eax*4]		;1st solution
	xor	eax,L[ecx*4]
	xor	edx,eax			;combined solution
	xor	edx,[edi]
	mov	[edi],edx
	add	edi,4
	sub	ebx,4			;adjust src/dst count, zero if done
	jnz	lp4
ewrap:
;--	mov	ebp,SFRM		;qword processing/alignment issues may
	mov	eax,edi			;  cause us to process a single dword
	sub	eax,outputbuffer	;  at the beginning and/or end.
	mov	ebx,bsize		;  Calc amount of unprocessed data.
	sub	ebx,eax			;ebx= bsize remaining to process
	jnz	chk8			;jmp= more data to process
eempt:
	emms
	add	esp,800h		;discard tables
	pop	ebp
	pop	esp
  }
}
